# -*- coding: utf-8 -*-

# Copyright (c) 2014-2017, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import logging
import httplib2
import copy
from StringIO import StringIO
from urlparse import urlsplit, urljoin
from urllib2 import urlopen

import xml.sax.handler
from xml.sax import saxutils
from xml.sax.saxutils import XMLFilterBase, XMLGenerator
from xml.sax.xmlreader import InputSource
import defusedxml.expatreader

from pyramid.httpexceptions import HTTPBadGateway

from owslib.wms import WebMapService

from c2cgeoportal.lib import caching, get_protected_layers_query, \
    get_writable_layers_query, add_url_params, get_ogc_server_wms_url_ids,\
    get_ogc_server_wfs_url_ids
from c2cgeoportal.models import DBSession, LayerWMS, OGCServer

cache_region = caching.get_region()
log = logging.getLogger(__name__)


@cache_region.cache_on_arguments()
def get_protected_layers(role_id, ogc_server_ids):
    q = get_protected_layers_query(role_id, ogc_server_ids, what=LayerWMS, version=2)
    results = q.all()
    DBSession.expunge_all()
    return {r.id: r for r in results}


@cache_region.cache_on_arguments()
def get_private_layers(ogc_server_ids):
    q = DBSession.query(LayerWMS) \
        .filter(LayerWMS.public.is_(False)) \
        .join(LayerWMS.ogc_server) \
        .filter(OGCServer.id.in_(ogc_server_ids))
    results = q.all()
    DBSession.expunge_all()
    return {r.id: r for r in results}


@cache_region.cache_on_arguments()
def get_writable_layers(role_id, ogc_server_ids):
    q = get_writable_layers_query(role_id, ogc_server_ids)
    results = q.all()
    DBSession.expunge_all()
    return {r.id: r for r in results}


@cache_region.cache_on_arguments()
def wms_structure(wms_url, host):
    url = urlsplit(wms_url)
    wms_url = add_url_params(wms_url, {
        "SERVICE": "WMS",
        "VERSION": "1.1.1",
        "REQUEST": "GetCapabilities",
    })

    # Forward request to target (without Host Header)
    http = httplib2.Http()
    headers = dict()
    if url.hostname == "localhost" and host is not None:  # pragma: no cover
        headers["Host"] = host
    try:
        resp, content = http.request(wms_url, method="GET", headers=headers)
    except:  # pragma: no cover
        raise HTTPBadGateway("Unable to GetCapabilities from wms_url {0!s}".format(wms_url))

    if resp.status < 200 or resp.status >= 300:  # pragma: no cover
        raise HTTPBadGateway(
            "GetCapabilities from wms_url {0!s} return the error: {1:d} {2!s}".format(
                wms_url, resp.status, resp.reason
            )
        )

    try:
        wms = WebMapService(None, xml=content)
        result = {}

        def _fill(name, parent):
            if parent is None:
                return
            if parent.name not in result:
                result[parent.name] = []
            result[parent.name].append(name)
            _fill(name, parent.parent)

        for layer in wms.contents.values():
            _fill(layer.name, layer.parent)
        return result

    except AttributeError:  # pragma: no cover
        error = "WARNING! an error occurred while trying to " \
            "read the mapfile and recover the themes."
        error = "{0!s}\nurl: {1!s}\nxml:\n{2!s}".format(error, wms_url, content)
        log.exception(error)
        raise HTTPBadGateway(error)

    except SyntaxError:  # pragma: no cover
        error = "WARNING! an error occurred while trying to " \
            "read the mapfile and recover the themes."
        error = "{0!s}\nurl: {1!s}\nxml:\n{2!s}".format(error, wms_url, content)
        log.exception(error)
        raise HTTPBadGateway(error)


def enable_proxies(proxies):  # pragma: no cover
    old_prepare_input_source = saxutils.prepare_input_source

    def caching_prepare_input_source(source, base=None):
        if isinstance(source, InputSource):
            return source

        full_uri = urljoin(base or "", source)

        # Convert StringIO to string
        if hasattr(full_uri, "getvalue"):
            full_uri = full_uri.getvalue()

        if not full_uri.startswith("http:"):
            args = (source,) if base is None else (source, base)
            return old_prepare_input_source(*args)

        document = urlopen(full_uri, proxies=proxies)

        input_source = InputSource()
        input_source.setSystemId(source)
        input_source.setByteStream(document)

        return input_source

    saxutils.prepare_input_source = caching_prepare_input_source


def filter_capabilities(content, role_id, wms, url, headers, proxies, request):

    if proxies:  # pragma: no cover
        enable_proxies(proxies)

    wms_structure_ = wms_structure(url, headers.get("Host"))

    ogc_server_ids = (
        get_ogc_server_wms_url_ids(request) if wms else
        get_ogc_server_wfs_url_ids(request)
    ).get(url)
    gmf_private_layers = copy.copy(get_private_layers(ogc_server_ids))
    for id_ in get_protected_layers(role_id, ogc_server_ids).keys():
        if id_ in gmf_private_layers:
            del gmf_private_layers[id_]

    private_layers = set()
    for gmflayer in gmf_private_layers.values():
        for ogclayer in gmflayer.layer.split(","):
            private_layers.add(ogclayer)
            if ogclayer in wms_structure_:
                private_layers.update(wms_structure_[ogclayer])

    parser = defusedxml.expatreader.create_parser(forbid_external=False)
    # skip inclusion of DTDs
    parser.setFeature(xml.sax.handler.feature_external_ges, False)
    parser.setFeature(xml.sax.handler.feature_external_pes, False)

    result = StringIO()
    downstream_handler = XMLGenerator(result, "utf-8")
    filter_handler = _CapabilitiesFilter(
        parser, downstream_handler,
        u"Layer" if wms else u"FeatureType",
        layers_blacklist=private_layers
    )
    filter_handler.parse(StringIO(content))
    return unicode(result.getvalue(), "utf-8")


def filter_wfst_capabilities(content, role_id, wfs_url, proxies, request):

    if proxies:  # pragma: no cover
        enable_proxies(proxies)

    writable_layers = []
    ogc_server_ids = get_ogc_server_wfs_url_ids(request).get(wfs_url)
    for gmflayer in get_writable_layers(role_id, ogc_server_ids).values():
        writable_layers += gmflayer.layer.split(",")

    parser = defusedxml.expatreader.create_parser(forbid_external=False)
    # skip inclusion of DTDs
    parser.setFeature(xml.sax.handler.feature_external_ges, False)
    parser.setFeature(xml.sax.handler.feature_external_pes, False)

    result = StringIO()
    downstream_handler = XMLGenerator(result, "utf-8")
    filter_handler = _CapabilitiesFilter(
        parser, downstream_handler,
        u"FeatureType",
        layers_whitelist=writable_layers
    )
    filter_handler.parse(StringIO(content))
    filtered_content = unicode(result.getvalue(), "utf-8")
    return filtered_content


class _Layer:
    def __init__(self, self_hidden=False):
        self.accumul = []
        self.hidden = True
        self.self_hidden = self_hidden
        self.has_children = False
        self.children_nb = 0


class _CapabilitiesFilter(XMLFilterBase):
    """
    SAX filter to show only the allowed layers in a GetCapabilities request.
    The filter removes elements of type `tag_name` where the `name`
    attribute is part of the set `layers_blacklist` (when `layers_blacklist`
    is given) or is not part of the set `layers_whitelist` (when
    `layers_whitelist` is given).
    """

    def __init__(
            self, upstream, downstream, tag_name,
            layers_blacklist=None, layers_whitelist=None):
        XMLFilterBase.__init__(self, upstream)
        self._downstream = downstream
        self._accumulator = []

        assert layers_blacklist is not None or layers_whitelist is not None, \
            "either layers_blacklist OR layers_whitelist must be set"
        assert not (
            layers_blacklist is not None and
            layers_whitelist is not None), \
            "only either layers_blacklist OR layers_whitelist can be set"

        if layers_blacklist is not None:
            layers_blacklist = [layer.lower() for layer in layers_blacklist]
        if layers_whitelist is not None:
            layers_whitelist = [layer.lower() for layer in layers_whitelist]
        self.layers_blacklist = layers_blacklist
        self.layers_whitelist = layers_whitelist

        self.layers_path = []
        self.in_name = False
        self.tag_name = tag_name
        self.level = 0

    def _complete_text_node(self):
        if self._accumulator:
            self._downstream.characters("".join(self._accumulator))
            self._accumulator = []

    def _do(self, action):
        if len(self.layers_path) != 0:
            self.layers_path[-1].accumul.append(action)
        else:
            self._complete_text_node()
            action()

    def _add_child(self, layer):
        if not layer.hidden and not (
                layer.has_children and layer.children_nb == 0
        ):
            for action in layer.accumul:
                self._complete_text_node()
                action()
            layer.accumul = []

    def setDocumentLocator(self, locator):  # noqa
        self._downstream.setDocumentLocator(locator)

    def startDocument(self):  # noqa
        self._downstream.startDocument()

    def endDocument(self):  # noqa
        self._downstream.endDocument()

    def startPrefixMapping(self, prefix, uri):  # pragma: no cover  # noqa
        self._downstream.startPrefixMapping(prefix, uri)

    def endPrefixMapping(self, prefix):  # pragma: no cover  # noqa
        self._downstream.endPrefixMapping(prefix)

    def startElement(self, name, attrs):  # noqa
        if name == self.tag_name:
            self.level += 1
            parent_layer = None
            if len(self.layers_path) > 0:
                parent_layer = self.layers_path[-1]
                parent_layer.has_children = True
                parent_layer.children_nb += 1
            layer = _Layer(
                parent_layer.self_hidden
            ) if len(self.layers_path) > 1 else _Layer()
            self.layers_path.append(layer)

            if parent_layer is not None:
                parent_layer.accumul.append(
                    lambda: self._add_child(layer)
                )
        elif name == "Name" and len(self.layers_path) != 0:
            self.in_name = True

        self._do(lambda: self._downstream.startElement(name, attrs))

    def endElement(self, name):  # noqa
        self._do(lambda: self._downstream.endElement(name))

        if name == self.tag_name:
            self.level -= 1
            if self.level == 0 and not self.layers_path[0].hidden:
                for action in self.layers_path[0].accumul:
                    self._complete_text_node()
                    action()
            self.layers_path.pop()
        elif name == "Name":
            self.in_name = False

    def startElementNS(self, name, qname, attrs):  # pragma: no cover  # noqa
        self._do(lambda: self._downstream.startElementNS(name, qname, attrs))

    def endElementNS(self, name, qname):  # pragma: no cover  # noqa
        self._do(lambda: self._downstream.endElementNS(name, qname))

    def _keep_layer(self, layer_name):
        return (
            self.layers_blacklist is not None and
            layer_name not in self.layers_blacklist) or (
            self.layers_whitelist is not None and
            layer_name in self.layers_whitelist)

    def characters(self, content):
        if self.in_name and len(self.layers_path) != 0 and not \
                self.layers_path[-1].self_hidden is True:
            layer_name = normalize_typename(content)
            if self._keep_layer(layer_name):
                for layer in self.layers_path:
                    layer.hidden = False
            else:
                # remove layer
                self.layers_path[-1].self_hidden = True
                if len(self.layers_path) > 1:
                    self.layers_path[-2].children_nb -= 1

        self._do(lambda: self._accumulator.append(content))

    def ignorableWhitespace(self, chars):  # pragma: no cover  # noqa
        self._do(lambda: self._accumulator.append(chars))

    def processingInstruction(self, target, data):  # pragma: no cover  # noqa
        self._do(lambda: self._downstream.processingInstruction(target, data))

    def skippedEntity(self, name):  # pragma: no cover  # noqa
        self._downstream.skippedEntity(name)


def normalize_tag(tag):
    """
    Drops the namespace from a tag and converts to lower case.
    e.g. '{http://....}TypeName' -> 'TypeName'
    """
    normalized = tag
    if len(tag) >= 3:
        if tag[0] == "{":
            normalized = tag[1:].split("}")[1]
    return normalized.lower()


def normalize_typename(typename):
    """
    Drops the namespace from a type name and converts to lower case.
    e.g. 'tows:parks' -> 'parks'
    """
    normalized = typename
    if ":" in typename:
        normalized = typename.split(":")[1]
    return normalized.lower()



# -*- coding: utf-8 -*-

# Copyright (c) 2016-2017, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

"""Add layer column in layerv1 table

Revision ID: 7d271f4527cd
Revises: 8117bb9bba16
Create Date: 2016-10-20 15:00:13.619090
"""

from alembic import op, context
from sqlalchemy import Column
from sqlalchemy.types import Unicode

# revision identifiers, used by Alembic.
revision = '7d271f4527cd'
down_revision = '8117bb9bba16'
branch_labels = None
depends_on = None


def upgrade():
    schema = context.get_context().config.get_main_option('schema')

    op.add_column('layerv1', Column('layer', Unicode), schema=schema)
    op.execute(
        'UPDATE {schema}.layerv1 AS l1 '
        'SET layer = name '
        'FROM {schema}.treeitem AS ti '
        'WHERE l1.id = ti.id'.format(schema=schema)
    )


def downgrade():
    schema = context.get_context().config.get_main_option('schema')

    op.drop_column('layerv1', 'layer', schema=schema)



# -*- coding: utf-8 -*-

# Copyright (c) 2012-2017, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import re
import subprocess
import os
import json
import requests
import yaml
from six import string_types
from simplejson.scanner import JSONDecodeError

from pyramid.scaffolds.template import Template
from pyramid.compat import input_
from c2cgeoportal.lib.bashcolor import colorize, GREEN


class BaseTemplate(Template):  # pragma: no cover
    """
    A class that can be used as a base class for c2cgeoportal scaffolding
    templates.

    Greatly inspired from ``pyramid.scaffolds.template.PyramidTemplate``.
    """

    def pre(self, command, output_dir, vars_):
        """
        Overrides ``pyramid.scaffold.template.Template.pre``, adding
        several variables to the default variables list. Also prevents
        common misnamings (such as naming a package "site" or naming a
        package logger "root").
        """

        self._get_vars(vars_, "package", "Get a package name: ")
        self._get_vars(vars_, "apache_vhost", "The Apache vhost name: ")
        self._get_vars(
            vars_, "srid",
            "Spatial Reference System Identifier (e.g. 21781): ", int,
        )
        srid = vars_["srid"]
        extent = self._epsg2bbox(srid)
        self._get_vars(
            vars_, "extent",
            "Extent (minx miny maxx maxy): in EPSG: {srid} projection, default is "
            "[{bbox[0]} {bbox[1]} {bbox[2]} {bbox[3]}]: ".format(srid=srid, bbox=extent)
            if extent else
            "Extent (minx miny maxx maxy): in EPSG: {srid} projection: ".format(srid=srid)
        )
        match = re.match(r"([\d.]+)[,; ] *([\d.]+)[,; ] *([\d.]+)[,; ] *([\d.]+)", vars_["extent"])
        if match is not None:
            extent = [match.group(n + 1) for n in range(4)]
        vars_["extent"] = ",".join(extent)
        vars_["extent_mapserver"] = " ".join(extent)
        vars_["extent_viewer"] = json.dumps(extent)

        ret = Template.pre(self, command, output_dir, vars_)

        if vars_["package"] == "site":
            raise ValueError(
                "Sorry, you may not name your package 'site'. "
                "The package name 'site' has a special meaning in "
                "Python.  Please name it anything except 'site'.")

        package_logger = vars_["package"]
        if package_logger == "root":
            # Rename the app logger in the rare case a project
            # is named "root"
            package_logger = "app"
        vars_["package_logger"] = package_logger

        return ret

    @staticmethod
    def out(msg):
        print(msg)

    @staticmethod
    def _get_vars(vars_, name, prompt, type_=None):
        """
        Set an attribute in the vars dict.
        """

        if name.upper() in os.environ:
            value = os.environ[name.upper()]
        else:
            value = vars_.get(name)

        if value is None:
            value = input_(prompt).strip()

        if type_ is not None:
            try:
                type_(value)
            except ValueError:
                print("The attribute {} is not a {}".format(name, type_))
                exit(1)

        vars_[name] = value

    @staticmethod
    def _epsg2bbox(srid):
        try:
            r = requests.get("http://epsg.io/?format=json&q={}".format(srid))
            bbox = r.json()["results"][0]["bbox"]
            r = requests.get(
                "http://epsg.io/trans?s_srs=4326&t_srs={srid}&data={bbox[1]},{bbox[0]}"
                        .format(srid=srid, bbox=bbox)
            )
            r1 = r.json()[0]
            r = requests.get(
                "http://epsg.io/trans?s_srs=4326&t_srs={srid}&data={bbox[3]},{bbox[2]}"
                        .format(srid=srid, bbox=bbox)
            )
            r2 = r.json()[0]
            return [r1["x"], r2["y"], r2["x"], r1["y"]]
        except JSONDecodeError:
            print("epsg.io doesn't return a correct json.")
            return None
        except IndexError:
            print("Unable to get the bbox")
            return None


class TemplateCreate(BaseTemplate):  # pragma: no cover
    _template_dir = "create"
    summary = "Template used to create a c2cgeoportal project"

    def post(self, command, output_dir, vars_):
        """
        Overrides the base template class to print the next step.
        """

        if os.name == 'posix':
            for file in ("post-restore-code", "pre-restore-database.mako"):
                dest = os.path.join(output_dir, "deploy/hooks", file)
                subprocess.check_call(["chmod", "+x", dest])

        self.out("\nContinue with:")
        self.out(colorize(
            ".build/venv/bin/pcreate -s c2cgeoportal_update ../{vars[project]} "
            "package={vars[package]} srid={vars[srid]} extent={vars[extent]}".format(vars=vars_),
            GREEN
        ))

        return BaseTemplate.post(self, command, output_dir, vars_)


class TemplateUpdate(BaseTemplate):  # pragma: no cover
    _template_dir = "update"
    summary = "Template used to update a c2cgeoportal project"

    def pre(self, command, output_dir, vars_):
        """
        Overrides the base template
        """

        if os.path.exists("project.yaml"):
            with open("project.yaml", "r") as f:
                project = yaml.safe_load(f)
                if "template_vars" in project:
                    for key, value in project["template_vars"].items():
                        vars_[key] = \
                            value.encode("utf-8") \
                            if isinstance(value, string_types) \
                            else value

        return BaseTemplate.pre(self, command, output_dir, vars_)

    def post(self, command, output_dir, vars_):
        """
        Overrides the base template class to print "Welcome to c2cgeoportal!"
        after a successful scaffolding rendering.
        """

        if os.name == 'posix':
            dest = os.path.join(output_dir, ".whiskey/action_hooks/pre-build.mako")
            subprocess.check_call(["chmod", "+x", dest])
        self.out(colorize("\nWelcome to c2cgeoportal!", GREEN))

        return BaseTemplate.post(self, command, output_dir, vars_)



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2017, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import logging
import os

from fanstatic import Library, Group, Resource
from fanstatic.core import set_resource_file_existence_checking
from pyramid_formalchemy.utils import TemplateEngine
from formalchemy import config as fa_config
from formalchemy import FieldSet, Grid, helpers
from formalchemy.fields import Field, FieldRenderer, CheckBoxSet, \
    SelectFieldRenderer, AttributeField
from formalchemy.validators import ValidationError
from formalchemy.helpers import password_field
from geoalchemy2 import Geometry
from mako.lookup import TemplateLookup
from fa.jquery import renderers as fa_renderers
from fa.jquery import fanstatic_resources
from pyramid_formalchemy import events as fa_events
from geoformalchemy.base import GeometryFieldRenderer
from pyramid.i18n import TranslationStringFactory
from sqlalchemy.orm.attributes import manager_of_class

from c2cgeoportal import models
from c2cgeoportal import (
    formalchemy_default_zoom,
    formalchemy_default_x, formalchemy_default_y,
    formalchemy_available_functionalities,
    formalchemy_available_metadata,
)

__all__ = [
    "Functionality", "User", "Role", "LayerGroup", "Theme",
    "LayerV1", "OGCServer", "LayerWMS",
    "LayerWMTS", "RestrictionArea", "Interface", "Metadata",
    "Dimension", "LayerV1Grid", "LayerGroupGrid", "ThemeGrid",
    "LayerWMTSGrid", "FunctionalityGrid", "RestrictionAreaGrid", "RoleGrid",
    "UserGrid", "InterfaceGrid", "MetadataGrid", "DimensionGrid"
]


log = logging.getLogger(__name__)
_ = TranslationStringFactory("c2cgeoportal")

fa_config.encoding = "utf-8"
fa_config.engine = TemplateEngine()

fanstatic_lib = Library("admin", "static")
admin_js = Resource(
    fanstatic_lib,
    "build/admin/admin.js",
    depends=[fanstatic_resources.jqueryui])
admin_css = Resource(
    fanstatic_lib,
    "build/admin/admin.css",
    depends=[fanstatic_resources.fa_uiadmin_css])

# HACK to invoke fanstatic to inject a script which content is dynamic:
# the content of the script sets OpenLayers.ImgPath to an url that is
# dynamically generated using request.static_url
olimgpath_js = None
# deactive resource checking in fanstatic
set_resource_file_existence_checking(False)


def get_fanstatic_resources(request):  # pragma: no cover
    global olimgpath_js
    olimgpath = request.static_url("c2cgeoportal:static/lib/openlayers/img/")

    def olimgpath_renderer(url):
        return u'<script>OpenLayers.ImgPath="{0!s}";</script>'.format(olimgpath)
    if olimgpath_js is None:
        olimgpath_js = Resource(
            fanstatic_lib,
            "whatever.js",
            renderer=olimgpath_renderer,
            depends=[admin_js])

    return Group([admin_js, olimgpath_js, admin_css])
# end of HACK


@fa_events.subscriber([models.Functionality, fa_events.IBeforeRenderEvent])
def before_render_functionnality(context, event):  # pragma: no cover
    get_fanstatic_resources(event.request).need()


@fa_events.subscriber([models.Theme, fa_events.IBeforeRenderEvent])
def before_render_theme(context, event):  # pragma: no cover
    get_fanstatic_resources(event.request).need()


@fa_events.subscriber([models.Layer, fa_events.IBeforeRenderEvent])
def before_render_layer(context, event):  # pragma: no cover
    get_fanstatic_resources(event.request).need()


@fa_events.subscriber([models.LayerGroup, fa_events.IBeforeRenderEvent])
def before_render_layergroup(context, event):  # pragma: no cover
    get_fanstatic_resources(event.request).need()


@fa_events.subscriber([models.RestrictionArea, fa_events.IBeforeRenderEvent])
def before_render_restrictionarea(context, event):  # pragma: no cover
    get_fanstatic_resources(event.request).need()


@fa_events.subscriber([models.Role, fa_events.IBeforeRenderEvent])
def before_render_role(context, event):  # pragma: no cover
    get_fanstatic_resources(event.request).need()


@fa_events.subscriber([models.User, fa_events.IBeforeRenderEvent])
def before_render_user(context, event):  # pragma: no cover
    get_fanstatic_resources(event.request).need()


# validator to check uniqueness of unique key in db (prevent duplicate key error)
def unique_validator(value, f):  # pragma: no cover
    query = f.query(f.model.__class__)
    query = query.filter_by(**{f.name: f.value})
    if f.parent._bound_pk is None and query.first():
        raise ValidationError(_(u"Duplicate record"))


class PyramidGeometryFieldRenderer(GeometryFieldRenderer):  # pragma: no cover
    def __init__(self, field):
        self.__templates = None
        GeometryFieldRenderer.__init__(self, field)

    def get_templates(self):
        if self.__templates is None:
            self.__templates = TemplateLookup(
                [os.path.join(os.path.dirname(__file__), "templates", "admin")],
                input_encoding="utf-8", output_encoding="utf-8")
        return self.__templates


FieldSet.default_renderers.update(fa_renderers.default_renderers)
FieldSet.default_renderers[Geometry] = PyramidGeometryFieldRenderer


class DblPasswordField(Field):  # pragma: no cover
    def __init__(self, parent, original):
        self._original = original
        Field.__init__(self, name=original.name, value=original.value)
        self.parent = parent

        def passwords_match(value, field):
            value1 = field.renderer.params.getone(field.renderer.name)
            value2 = field.renderer.params.getone(field.renderer.name + "_confirm")
            if value1 != value2:
                raise ValidationError(_("Passwords do not match"))
        self.validators.append(passwords_match)

    def sync(self):
        value = self.renderer.params.getone(self.renderer.name)
        if len(value) > 0:
            setattr(self.model, self.name, value)

    def render(self):
        return (
            password_field(self.renderer.name, value="") +
            password_field(self.renderer.name + "_confirm", value=""))


class CheckBoxTreeSet(CheckBoxSet):  # pragma: no cover
    def __init__(self, attribute, dom_id, auto_check=True, auto_collapsed=True):
        super(CheckBoxTreeSet, self).__init__(attribute)
        self.dom_id = dom_id
        self.auto_check = auto_check
        self.auto_collapsed = auto_collapsed

    @staticmethod
    def render_tree():
        return ""

    def render(self, options, **kwargs):
        opt = u""
        if self.auto_collapsed:
            opt += u'"initializeUnchecked": "collapsed"'
        if self.auto_collapsed and not self.auto_check:
            opt += u","
        if not self.auto_check:
            opt += u"""'onCheck': {
                'others': false,
                'descendants': false,
                'ancestors': false },
            'onUncheck': {
                'others': false,
                'descendants': false,
                'ancestors': false }"""
        result = u"""<script lang="text/javascript" >
            $(document).ready(function(){{
                $("#{id!s}").checkboxTree({{{opt!s}}});
            }});
        </script>
        <ul id="{id!s}" class="checkboxtree">
        """.format(id=self.dom_id, opt=opt)
        result += self.render_tree()
        result += u"</ul>"
        return result


class SimpleLayerCheckBoxTreeSet(CheckBoxTreeSet):  # pragma: no cover

    def __init__(
            self, attribute, dom_id="layer_tree",
            auto_check=True, only_internal_wms=True):
        super(SimpleLayerCheckBoxTreeSet, self).__init__(attribute, dom_id, auto_check)
        self._rendered_id = []
        self.only_internal_wms = only_internal_wms

    def render_children(self, item, depth):
        # escape loop
        if (depth >= 5):
            return ""

        result = u""
        if isinstance(item, models.TreeGroup):
            result += u"<ul>"
            for child in item.children_relation:
                result += self.render_item(child, depth + 1)
            result += u"</ul>"
        return result

    def render_organisational_item(self, item, depth):
        if item in self.layer_group:
            self.layer_group.remove(item)

        result = u"<li>"
        if self.auto_check:
            result += u'<input type="checkbox"></input>'
        result += u"<label>{label!s}</label>".format(
            label=item.name,
        )
        result += self.render_children(item, depth)
        result += u"</li>"
        return result

    def is_checked(self, item, final_item):
        return final_item.id in self.values

    def build_values(self):
        self.values = [int(v) for v in self.value]

    def render_item(self, item, depth):
        final_item = item.treeitem if isinstance(item, models.LayergroupTreeitem) else item

        # no link to theme
        # if autocheck mean that we want select only layers.
        if isinstance(final_item, models.Theme) or \
                self.auto_check and not isinstance(final_item, models.Layer):
            return self.render_organisational_item(final_item, depth)

        # escape public layer if wanted
        if self.only_internal_wms and isinstance(final_item, models.LayerV1) \
                and final_item.layer_type != "internal WMS":
            return ""

        if final_item in self.layer_group:
            self.layer_group.remove(final_item)
        elif final_item in self.layer:
            self.layer.remove(final_item)

        prefixs = {
            "layerv1": "Layer V1: ",
            "l_wms": "WMS Layer: ",
            "l_wmts": "WMTS Layer: ",
        }

        result = u"""
        <li>
            <input type="checkbox" id="{id!s}" name="{name!s}" value="{value!s}"{add!s}></input>
            <label>{type!s}{label!s}</label>
            """.format(
            id="{}_{}".format(self.name, self.i),
            # adds -second to fields (layer) that appears two time to
            # don"t save them twice (=> integrity error).
            name=self.name + ("-second" if final_item.id in self._rendered_id else ""),
            value=final_item.id,
            add=' checked="checked"' if self.is_checked(item, final_item) else "",
            type=prefixs.get(final_item.item_type, ""),
            label=final_item.name,
        )
        self._rendered_id.append(final_item.id)
        self.i += 1
        result += self.render_children(final_item, depth)
        result += u"</li>"
        return result

    def render_tree(self):
        self.build_values()

        self.layer = models.DBSession.query(models.Layer).all()
        self.layer_group = models.DBSession.query(models.LayerGroup).all()
        themes = models.DBSession.query(models.Theme). \
            order_by(models.Theme.ordering).all()
        self.i = 0
        result = u""
        for item in themes:
            result += self.render_item(item, 1)

        # add unlinked layers
        if len(self.layer) >= 0 or len(self.layer_group) > 0:
            result += u"<li>"
            if self.auto_check:
                result += u'<input type="checkbox"></input>'
            result += u"<label>{name!s}</label>".format(name=_("Unlinked layers"))
            result += u"<ul>"

            while len(self.layer_group) > 0:
                result += self.render_item(self.layer_group.pop(0), 2)
            while len(self.layer) > 0:
                result += self.render_item(self.layer.pop(0), 2)

            result += u"</ul>"
            result += u"</li>"
        return result


class LayerCheckBoxTreeSet(SimpleLayerCheckBoxTreeSet):  # pragma: no cover
    def build_values(self):
        values = models.DBSession.query(models.LayergroupTreeitem) \
            .filter(models.LayergroupTreeitem.id.in_([int(v) for v in self.value])).all()
        self.values = [v.treeitem.id for v in values]


class TreeItemCheckBoxTreeSet(LayerCheckBoxTreeSet):  # pragma: no cover
    def __init__(self, attribute):
        super(TreeItemCheckBoxTreeSet, self).__init__(
            attribute,
            auto_check=False, only_internal_wms=False
        )


class FunctionalityCheckBoxTreeSet(CheckBoxTreeSet):  # pragma: no cover
    def __init__(self, attribute):
        super(FunctionalityCheckBoxTreeSet, self).__init__(
            attribute, dom_id="tree_func", auto_collapsed=False)

    def render_tree(self):
        query = models.DBSession.query(models.Functionality)
        query = query.order_by(models.Functionality.name)
        query = query.order_by(models.Functionality.value)
        functionalities = query.all()
        i = 0
        prev_name = u""
        result = u""
        for functionality in functionalities:
            if prev_name != functionality.name:
                if prev_name != u"":
                    result += u"</ul></li>\n"
                prev_name = functionality.name
                result += \
                    u'<li><input type="checkbox" style="display:none"></input>' \
                    '<label>%s</label><ul>\n' % (functionality.name)
            result += \
                u'<li><input type="checkbox" id="%s" name="%s" value="%i"%s>' \
                '</input><label>%s</label></li>\n' % (
                    "{0!s}_{1:d}".format(self.name, i),
                    self.name,
                    functionality.id,
                    u' checked="checked"' if self._is_checked(functionality.id) else "",
                    functionality.value)
            i += 1
        result += u"</ul></li>"
        return result


class RoListRenderer(FieldRenderer):  # pragma: no cover
    def render_readonly(self, **kwargs):
        return helpers.content_tag(u"span", ("," + helpers.tag("br")).join([
            helpers.literal(value) for value in self.raw_value
        ]), style=u"white-space: nowrap;")


##############################################################################
# FIELDS defs
#
# DefaultBasemap, Layer, LayerGroup, Mandant, Printtemplates, RestrictionArea,
# Role, Title, User
#
##############################################################################

image_type_options = [
    (_("image/jpeg"), "image/jpeg"),
    (_("image/png"), "image/png")
]
time_options = [
    (_("disabled"), "disabled"),
    (_("value"), "value"),
    (_("range"), "range"),
]
time_widget_options = [
    (_("slider"), "slider"),
    (_("datepicker"), "datepicker"),
]
ogcserver_type_options = [
    (_("MapServer"), models.OGCSERVER_TYPE_MAPSERVER),
    (_("QGISserver"), models.OGCSERVER_TYPE_QGISSERVER),
    (_("GeoServer"), models.OGCSERVER_TYPE_GEOSERVER),
    (_("Other"), models.OGCSERVER_TYPE_OTHER),
]
ogcserver_auth_options = [
    (_("No auth"), models.OGCSERVER_AUTH_NOAUTH),
    (_("Standard auth"), models.OGCSERVER_AUTH_STANDARD),
    (_("GeoServer auth"), models.OGCSERVER_AUTH_GEOSERVER),
]

# Layer V1
LayerV1 = FieldSet(models.LayerV1)
LayerV1.configure(exclude=[LayerV1.parents_relation])
LayerV1.layer_type.set(
    renderer=SelectFieldRenderer,
    options=[
        (_("internal WMS"), "internal WMS"),
        (_("external WMS"), "external WMS"),
        (_("WMTS"), "WMTS"),
        (_("no 2D"), "no 2D"),
    ])
LayerV1.image_type.set(
    renderer=SelectFieldRenderer,
    options=image_type_options
)
LayerV1.time_mode.set(
    renderer=SelectFieldRenderer,
    options=time_options,
)
LayerV1.time_widget.set(
    renderer=SelectFieldRenderer,
    options=time_widget_options,
)
LayerV1.interfaces.set(renderer=CheckBoxSet)
LayerV1.metadatas.set(renderer=RoListRenderer, readonly=True)
LayerV1.restrictionareas.set(renderer=CheckBoxSet)

# OGC server
OGCServer = FieldSet(models.OGCServer)
OGCServer.image_type.set(
    renderer=SelectFieldRenderer,
    options=image_type_options,
)
OGCServer.type.set(
    renderer=SelectFieldRenderer,
    options=ogcserver_type_options,
)
OGCServer.auth.set(
    renderer=SelectFieldRenderer,
    options=ogcserver_auth_options,
)

# LayerWMS
LayerWMS = FieldSet(models.LayerWMS)
LayerWMS.configure(exclude=[LayerWMS.parents_relation])
LayerWMS.time_mode.set(
    renderer=SelectFieldRenderer,
    options=time_options,
)
LayerWMS.time_widget.set(
    renderer=SelectFieldRenderer,
    options=time_widget_options,
)
LayerWMS.interfaces.set(renderer=CheckBoxSet)
LayerWMS.metadatas.set(renderer=RoListRenderer, readonly=True)
LayerWMS.dimensions.set(renderer=RoListRenderer, readonly=True)
LayerWMS.restrictionareas.set(renderer=CheckBoxSet)
LayerWMS.ogc_server.set(renderer=SelectFieldRenderer)


# LayerWMTS
LayerWMTS = FieldSet(models.LayerWMTS)
LayerWMTS.configure(exclude=[LayerWMTS.parents_relation])
LayerWMTS.image_type.set(
    renderer=SelectFieldRenderer,
    options=image_type_options
)
LayerWMTS.interfaces.set(renderer=CheckBoxSet)
LayerWMTS.metadatas.set(renderer=RoListRenderer, readonly=True)
LayerWMTS.dimensions.set(renderer=RoListRenderer, readonly=True)
LayerWMTS.restrictionareas.set(renderer=CheckBoxSet)


class ChildrenAttributeField(AttributeField):

    def __init__(self, *args, **kargs):
        AttributeField.__init__(self, *args, **kargs)

    def sync(self):  # pragma: no cover
        self.model.children = [
            self.parent.session.query(models.TreeItem).get(int(pk))
            for pk in self.renderer.deserialize()
        ]


# LayerGroup
LayerGroup = FieldSet(models.LayerGroup)
LayerGroup.configure(exclude=[LayerGroup.parents_relation, LayerGroup.children_relation])
LayerGroup.append(ChildrenAttributeField(
    manager_of_class(models.LayerGroup)["children_relation"], LayerGroup,
))
LayerGroup.children_relation.set(renderer=TreeItemCheckBoxTreeSet)
LayerGroup.metadatas.set(readonly=True)

# LayergroupTreeitem
LayergroupTreeitem = FieldSet(models.LayergroupTreeitem)
LayergroupTreeitem.ordering.set(metadata=dict(mandatory="")).required()

# Theme
Theme = FieldSet(models.Theme)
Theme.configure(exclude=[Theme.parents_relation, Theme.children_relation])
Theme.ordering.set(metadata=dict(mandatory="")).required()
Theme.append(ChildrenAttributeField(
    manager_of_class(models.Theme)["children_relation"], Theme
))
Theme.children_relation.set(renderer=TreeItemCheckBoxTreeSet)
Theme.functionalities.set(renderer=FunctionalityCheckBoxTreeSet)
Theme.interfaces.set(renderer=CheckBoxSet)
Theme.metadatas.set(readonly=True)
Theme.restricted_roles.set(renderer=CheckBoxSet)

# Functionality
Functionality = FieldSet(models.Functionality)
Functionality.name.set(
    renderer=SelectFieldRenderer,
    options=[(f, f) for f in formalchemy_available_functionalities])
Functionality.value.set(metadata=dict(mandatory="")).required()

# Interface
Interface = FieldSet(models.Interface)
Interface.configure(include=[Interface.name, Interface.description])

# Metadata
Metadata = FieldSet(models.Metadata)
Metadata.configure(include=[
    Metadata.item,
    Metadata.name,
    Metadata.value,
    Metadata.description
])
Metadata.name.set(
    renderer=SelectFieldRenderer,
    options=[(m, m) for m in formalchemy_available_metadata])
Metadata.value.set(metadata=dict(mandatory="")).required()

# Dimension
Dimension = FieldSet(models.Dimension)
Dimension.name.set(metadata=dict(mandatory="")).required()

# RestrictionArea
RestrictionArea = FieldSet(models.RestrictionArea)
RestrictionArea.name.set(metadata=dict(mandatory="")).required()
RestrictionArea.layers.set(renderer=SimpleLayerCheckBoxTreeSet)
RestrictionArea.roles.set(renderer=CheckBoxSet)
RestrictionArea.area.set(label=_(u"Restriction area"), options=[
    ("map_srid", 3857),
    ("base_layer", 'new OpenLayers.Layer.OSM("OSM")'),
    ("zoom", formalchemy_default_zoom),
    ("default_lon", formalchemy_default_x),
    ("default_lat", formalchemy_default_y)
])
field_order = [
    RestrictionArea.name,
    RestrictionArea.description,
    RestrictionArea.layers,
    RestrictionArea.roles,
    RestrictionArea.readwrite,
    RestrictionArea.area
]
RestrictionArea.configure(include=field_order)

# Role
Role = FieldSet(models.Role)
Role.name.set(metadata=dict(mandatory="")).required()
Role.functionalities.set(renderer=FunctionalityCheckBoxTreeSet)
Role.restrictionareas.set(renderer=CheckBoxSet)
Role.extent.set(label=_(u"Extent"), options=[
    ("map_srid", 3857),
    ("base_layer", 'new OpenLayers.Layer.OSM("OSM")'),
    ("zoom", formalchemy_default_zoom),
    ("default_lon", formalchemy_default_x),
    ("default_lat", formalchemy_default_y)
])
field_order = [
    Role.name,
    Role.description,
    Role.functionalities,
    Role.restrictionareas,
    Role.extent
]
Role.configure(include=field_order)

# User
User = FieldSet(models.User)


def _get_roles(parent):  # pragma: no cover
    return [
        (role.name, role.name)
        for role in parent.session.query(models.Role).all()
    ]


User.role_name.set(
    renderer=SelectFieldRenderer,
    options=_get_roles,
)
password = DblPasswordField(User, User._password)
User.append(password)
field_order = [
    User.username.validate(unique_validator).with_metadata(mandatory=""),
    password,
    User.role_name
]
if hasattr(User, "parent_role_name"):  # pragma: no cover
    field_order.append(User.parent_role_name)
field_order.extend([
    User.email.with_metadata(mandatory="")
])
User.configure(include=field_order)

#############################################################################
# GRID defs
#
# DefaultBasemapGrid, LayerGrid, LayerGroupGrid, MandantGrid,
# PrinttemplatesGrid, RestrictionAreaGrid, RoleGrid, TitleGrid, UserGrid
#
#############################################################################

# LayerV1Grid
LayerV1Grid = Grid(models.LayerV1)
field_order = [
    LayerV1.name,
    LayerV1.public,
    LayerV1.is_checked,
    LayerV1.icon,
    LayerV1.legend,
    LayerV1.identifier_attribute_field
]
LayerV1Grid.configure(include=field_order)

# LayerWMTSGrid
LayerWMTSGrid = Grid(models.LayerWMTS)
field_order = [
    LayerWMTS.name,
    LayerWMTS.public,
    LayerWMTS.layer
]
LayerWMTSGrid.configure(include=field_order)

# LayerGroupGrid
LayerGroupGrid = Grid(models.LayerGroup)
LayerGroupGrid.configure(exclude=[LayerGroupGrid.children_relation])

# LayergroupTreeitemGrid
LayergroupTreeitemGrid = Grid(models.LayergroupTreeitem)

# ThemeGrid
ThemeGrid = Grid(models.Theme)
ThemeGrid.configure(exclude=[ThemeGrid.parents_relation, ThemeGrid.children_relation])

# FunctionalityGrid
FunctionalityGrid = Grid(models.Functionality)

# InterfaceGrid
InterfaceGrid = Grid(models.Interface)

# MetadataGrid
MetadataGrid = Grid(models.Metadata)

# DimensionGrid
DimensionGrid = Grid(models.Dimension)

# RestrictionAreaGrid
RestrictionAreaGrid = Grid(models.RestrictionArea)
field_order = [
    RestrictionArea.name,
    RestrictionArea.description,
    RestrictionAreaGrid.roles
]
RestrictionAreaGrid.configure(include=field_order)

# RoleGrid
RoleGrid = Grid(models.Role)
field_order = [
    Role.name,
    Role.description,
    RoleGrid.functionalities,
    RoleGrid.restrictionareas,
]
RoleGrid.configure(include=field_order)

# UserGrid
UserGrid = Grid(models.User)
field_order = [User.username, User.role_name]
if hasattr(UserGrid, "parent_role"):  # pragma: no cover
    field_order.append(User.parent_role_name)
UserGrid.configure(include=field_order)



# -*- coding: utf-8 -*-

# Copyright (c) 2015-2017, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

"""internal and external layer tables refactoring, new ogc table

Revision ID: 116b9b79fc4d
Revises: 1418cb05921b
Create Date: 2015-10-28 12:21:59.162238
"""

from alembic import op, context
from sqlalchemy import ForeignKey, Column
from sqlalchemy.types import Integer, Boolean, Unicode

# revision identifiers, used by Alembic.
revision = '116b9b79fc4d'
down_revision = 'a4f1aac9bda'
branch_labels = None
depends_on = None


def upgrade():
    schema = context.get_context().config.get_main_option('schema')

    # Instructions
    op.create_table(
        'server_ogc',
        Column('id', Integer, primary_key=True),
        Column('name', Unicode, nullable=False),
        Column('description', Unicode),
        Column('url', Unicode),
        # url_wfs needed for Arcgis because wms and wfs url may be different
        Column('url_wfs', Unicode),
        Column('type', Unicode),
        Column('image_type', Unicode),
        Column('auth', Unicode),
        Column('wfs_support', Boolean, server_default='false'),
        Column('is_single_tile', Boolean, server_default='false'),
        schema=schema,
    )

    op.create_table(
        'layer_wms',
        Column(
            'id', Integer,
            ForeignKey(schema + '.layer.id'), primary_key=True
        ),
        Column(
            'server_ogc_id', Integer,
            ForeignKey(schema + '.server_ogc.id')
        ),
        Column('layer', Unicode),
        Column('style', Unicode),
        Column('time_mode', Unicode, server_default=u'disabled', nullable=False),
        Column('time_widget', Unicode, server_default=u'slider', nullable=False),
        schema=schema,
    )

    # move data from layer_internal_wms and layer_external_wms to the new
    # layer_wms and server_ogc tables

    # ocg for internal

    # default 'image/jpeg', 'image/png'
    op.execute(
        'INSERT INTO %(schema)s.server_ogc (name, description, type, image_type, '
        "  auth, wfs_support) "
        "SELECT 'source for ' || image_type AS name, "
        "  'default source for internal ' || image_type AS description, "
        "  'mapserver' AS type, "
        "  image_type, "
        "  'Standard auth' AS auth, "
        "  'true' AS wfs_support "
        "FROM ("
        "  SELECT UNNEST(ARRAY['image/jpeg', 'image/png']) AS image_type"
        ") AS foo" % {
            'schema': schema,
        }
    )
    # other custom image types
    op.execute(
        'INSERT INTO %(schema)s.server_ogc (name, description, type, image_type, '
        "  auth, wfs_support) "
        "SELECT 'source for ' || image_type AS name, "
        "  'default source for internal ' || image_type AS description, "
        "  'mapserver' AS type, "
        "  image_type, "
        "  'Standard auth' AS auth, "
        "  'true' AS wfs_support "
        "FROM ("
        "  SELECT DISTINCT(image_type) FROM %(schema)s.layer_internal_wms "
        "  WHERE image_type NOT IN ('image/jpeg', 'image/png')"
        ") as foo" % {
            'schema': schema,
        }
    )

    # layers for internal

    # internal with not null image_type
    op.execute(
        'INSERT INTO %(schema)s.layer_wms (id, server_ogc_id, layer, style, '
        '  time_mode, time_widget) '
        'SELECT lew.id, so.id, layer, style, time_mode, time_widget '
        'FROM %(schema)s.layer_internal_wms AS lew, %(schema)s.server_ogc AS so '
        'WHERE lew.image_type=so.image_type AND so.type IS NOT NULL' % {
            'schema': schema,
        }
    )
    # internal with null image_type
    op.execute(
        'INSERT INTO %(schema)s.layer_wms (id, server_ogc_id, layer, style, '
        '  time_mode, time_widget) '
        'SELECT lew.id, so.id, layer, style, time_mode, time_widget '
        'FROM %(schema)s.layer_internal_wms AS lew, %(schema)s.server_ogc AS so '
        "WHERE lew.image_type IS NULL AND so.image_type='image/png'" % {
            'schema': schema,
        }
    )

    # ocg for externals
    op.execute(
        'INSERT INTO %(schema)s.server_ogc (name, url, type, image_type, auth, is_single_tile) '
        "SELECT 'source for ' || url, url, 'mapserver' AS type, image_type, 'none', CASE "
        'WHEN is_single_tile IS TRUE THEN TRUE ELSE FALSE END as is_single_tile '
        'FROM %(schema)s.layer_external_wms GROUP BY url, image_type, is_single_tile' % {
            'schema': schema,
        }
    )

    # layers for external
    op.execute(
        'INSERT INTO %(schema)s.layer_wms (id, server_ogc_id, layer, style, '
        '  time_mode, time_widget) '
        'SELECT lew.id, so.id, layer, style, time_mode, time_widget '
        'FROM %(schema)s.layer_external_wms as lew, %(schema)s.server_ogc as so '
        'WHERE lew.url=so.url AND lew.is_single_tile=so.is_single_tile '
        'AND lew.image_type=so.image_type' % {
            'schema': schema,
        }
    )

    op.drop_table('layer_external_wms', schema=schema)
    op.drop_table('layer_internal_wms', schema=schema)

    # update layer type in treeitems
    op.execute(
        'UPDATE %(schema)s.treeitem '
        "SET type='l_wms' "
        "WHERE type='l_int_wms' OR type='l_ext_wms'" % {
            'schema': schema,
        }
    )


def downgrade():
    schema = context.get_context().config.get_main_option('schema')

    # Instructions

    # recreate tables 'layer_internal_wms' and 'layer_external_wms'
    op.create_table(
        'layer_internal_wms',
        Column(
            'id', Integer, ForeignKey(schema + '.layer.id'), primary_key=True
        ),
        Column('layer', Unicode),
        Column('image_type', Unicode(10)),
        Column('style', Unicode),
        Column('time_mode', Unicode(8)),
        Column('time_widget', Unicode(10), server_default=u'slider'),
        schema=schema,
    )

    op.create_table(
        'layer_external_wms',
        Column(
            'id', Integer, ForeignKey(schema + '.layer.id'), primary_key=True
        ),
        Column('url', Unicode),
        Column('layer', Unicode),
        Column('image_type', Unicode(10)),
        Column('style', Unicode),
        Column('is_single_tile', Boolean),
        Column('time_mode', Unicode(8)),
        Column('time_widget', Unicode(10), server_default=u'slider'),
        schema=schema,
    )
    # move data back

    # internal (type is not null)
    op.execute(
        'INSERT INTO %(schema)s.layer_internal_wms (id, layer, image_type, style, '
        '  time_mode, time_widget) '
        'SELECT w.id, layer, image_type, style, time_mode, time_widget '
        'FROM %(schema)s.layer_wms AS w, %(schema)s.server_ogc AS o '
        'WHERE w.server_ogc_id=o.id AND o.type IS NOT NULL' % {
            'schema': schema,
        }
    )

    # external (type is null)
    op.execute(
        'INSERT INTO %(schema)s.layer_external_wms (id, url, layer, image_type, style, '
        '  is_single_tile, time_mode, time_widget) '
        'SELECT w.id, url, layer, image_type, style, is_single_tile, time_mode, time_widget '
        'FROM %(schema)s.layer_wms AS w, %(schema)s.server_ogc AS o '
        'WHERE w.server_ogc_id=o.id AND o.type IS NULL' % {
            'schema': schema,
        }
    )

    # drop table AFTER moving data back
    op.drop_table('layer_wms', schema=schema)
    op.drop_table('server_ogc', schema=schema)

    # update layer type in treeitems
    # internal
    op.execute(
        "UPDATE %(schema)s.treeitem "
        "SET type='l_int_wms' "
        "FROM %(schema)s.layer_internal_wms as w "
        "WHERE %(schema)s.treeitem.id=w.id" % {
            'schema': schema,
        }
    )
    # external
    op.execute(
        "UPDATE %(schema)s.treeitem "
        "SET type='l_ext_wms' "
        "FROM %(schema)s.layer_external_wms as w "
        "WHERE %(schema)s.treeitem.id=w.id" % {
            'schema': schema,
        }
    )



# -*- coding: utf-8 -*-

# Copyright (c) 2011-2017, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.


import logging
from c2cgeoportal.lib import get_setting, get_typed, get_types_map

log = logging.getLogger(__name__)


def _get_config_functionality(name, registered, types, request, errors):
    result = None

    if registered:
        result = get_setting(
            request.registry.settings, ("functionalities", "registered", name))
    if result is None:
        result = get_setting(
            request.registry.settings, ("functionalities", "anonymous", name))

    if result is None:
        result = []
    elif not isinstance(result, list):
        result = [result]

    result = [get_typed(name, r, types, request, errors) for r in result]
    return [r for r in result if r is not None]


def _get_db_functionality(name, role, types, request, errors):
    result = [
        get_typed(name, functionality.value, types, request, errors)
        for functionality in role.functionalities
        if functionality.name == name
    ]
    return [r for r in result if r is not None]


FUNCTIONALITIES_TYPES = None


def get_functionality(name, request):
    global FUNCTIONALITIES_TYPES

    result = []
    errors = set()
    if FUNCTIONALITIES_TYPES is None:
        FUNCTIONALITIES_TYPES = get_types_map(
            request.registry.settings.get("admin_interface", {})
            .get("available_functionalities", [])
        )

    if request.user is not None and request.user.role is not None:
        result = _get_db_functionality(
            name, request.user.role,
            FUNCTIONALITIES_TYPES, request, errors
        )
    if len(result) == 0:
        result = _get_config_functionality(
            name, request.user is not None, FUNCTIONALITIES_TYPES, request, errors
        )

    if errors != set():  # pragma: no cover
        log.error("\n".join(errors))
    return result


def get_mapserver_substitution_params(request):
    params = {}
    mss = get_functionality(
        "mapserver_substitution", request
    )
    if mss:
        for s in mss:
            index = s.find("=")
            if index > 0:
                attribute = "s_" + s[:index]
                value = s[index + 1:]
                if attribute in params:
                    params[attribute] += "," + value
                else:
                    params[attribute] = value
            else:
                log.warning(
                    "Mapserver Substitution '%s' does not "
                    "respect pattern: <attribute>=<value>" % s
                )
    return params



# -*- coding: utf-8 -*-

# Copyright (c) 2014-2017, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

"""Add ordering in relation

Revision ID: 54645a535ad6
Revises: 415746eb9f6
Create Date: 2014-11-25 14:39:05.110315
"""

from alembic import op, context
from sqlalchemy import Column
from sqlalchemy.types import Integer

# revision identifiers, used by Alembic.
revision = '54645a535ad6'
down_revision = '415746eb9f6'


def upgrade():
    schema = context.get_context().config.get_main_option('schema')

    op.drop_constraint('layergroup_treeitem_pkey', 'layergroup_treeitem', schema=schema)
    op.add_column('layergroup_treeitem', Column('id', Integer, primary_key=True), schema=schema)
    op.add_column('layergroup_treeitem', Column('ordering', Integer), schema=schema)
    op.execute(
        'UPDATE ONLY %(schema)s.layergroup_treeitem AS lt SET ordering = ti."order" '
        'FROM %(schema)s.treeitem AS ti WHERE ti.id = lt.treeitem_id ' % {'schema': schema}
    )
    op.add_column('theme', Column('ordering', Integer), schema=schema)
    op.execute(
        'UPDATE ONLY %(schema)s.theme AS t SET ordering = ti."order" '
        'FROM %(schema)s.treeitem AS ti WHERE ti.id = t.id ' % {'schema': schema}
    )
    op.drop_column('treeitem', 'order', schema=schema)


def downgrade():
    schema = context.get_context().config.get_main_option('schema')
    op.add_column('treeitem', Column('order', Integer), schema=schema)
    op.execute(
        'UPDATE ONLY %(schema)s.treeitem AS ti SET "order" = lt.ordering '
        'FROM %(schema)s.layergroup_treeitem AS lt WHERE ti.id = lt.treeitem_id ' % {
            'schema': schema
        }
    )
    op.execute(
        'UPDATE ONLY %(schema)s.treeitem AS ti SET "order" = t.ordering '
        'FROM %(schema)s.theme AS t WHERE ti.id = t.id ' % {'schema': schema}
    )
    op.drop_column('theme', 'ordering', schema=schema)
    op.drop_column('layergroup_treeitem', 'ordering', schema=schema)
    op.drop_column('layergroup_treeitem', 'id', schema=schema)
    op.create_primary_key(
        'layergroup_treeitem_pkey', 'layergroup_treeitem',
        ['treegroup_id', 'treeitem_id'], schema=schema
    )



# -*- coding: utf-8 -*-

# Copyright (c) 2014-2017, Camptocamp SA
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

"""Changes to start the implementation of the version 2

Revision ID: 415746eb9f6
Revises: None
Create Date: 2014-10-23 16:00:47.940216
"""

from alembic import op, context
from sqlalchemy import Column, ForeignKey, Table, MetaData
from sqlalchemy.types import Integer, Boolean, Unicode, Float

# revision identifiers, used by Alembic.
revision = '415746eb9f6'
down_revision = '166ff2dcc48d'


def upgrade():
    schema = context.get_context().config.get_main_option('schema')

    engine = op.get_bind().engine
    if type(engine).__name__ != 'MockConnection' and \
            op.get_context().dialect.has_table(
                engine, 'interface', schema=schema):  # pragma: no cover
        return

    op.drop_table('user_functionality', schema=schema)

    op.create_table(
        'interface',
        Column(
            'id', Integer, primary_key=True
        ),
        Column('name', Unicode),
        Column('description', Unicode),
        schema=schema,
    )

    op.create_table(
        'interface_layer',
        Column(
            'interface_id', Integer, ForeignKey(schema + '.interface.id'), primary_key=True
        ),
        Column(
            'layer_id', Integer, ForeignKey(schema + '.layer.id'), primary_key=True
        ),
        schema=schema,
    )

    op.create_table(
        'interface_theme',
        Column(
            'interface_id', Integer, ForeignKey(schema + '.interface.id'), primary_key=True
        ),
        Column(
            'theme_id', Integer, ForeignKey(schema + '.theme.id'), primary_key=True
        ),
        schema=schema,
    )

    op.create_table(
        'layerv1',
        Column(
            'id', Integer, ForeignKey(schema + '.layer.id'), primary_key=True
        ),
        Column('is_checked', Boolean, default=True),
        Column('icon', Unicode),
        Column('layer_type', Unicode(12)),
        Column('url', Unicode),
        Column('image_type', Unicode(10)),
        Column('style', Unicode),
        Column('dimensions', Unicode),
        Column('matrix_set', Unicode),
        Column('wms_url', Unicode),
        Column('wms_layers', Unicode),
        Column('query_layers', Unicode),
        Column('kml', Unicode),
        Column('is_single_tile', Boolean),
        Column('legend', Boolean, default=True),
        Column('legend_image', Unicode),
        Column('legend_rule', Unicode),
        Column('is_legend_expanded', Boolean, default=False),
        Column('min_resolution', Float),
        Column('max_resolution', Float),
        Column('disclaimer', Unicode),
        Column('identifier_attribute_field', Unicode),
        Column('exclude_properties', Unicode),
        Column('time_mode', Unicode(8)),
        schema=schema,
    )

    op.execute(
        "UPDATE ONLY %(schema)s.treeitem SET type = 'layerv1' "
        "WHERE type='layer'" % {'schema': schema}
    )

    op.execute(
        'INSERT INTO %(schema)s.layerv1 ('
        'id, is_checked, icon, layer_type, url, image_type, style, dimensions, matrix_set, '
        'wms_url, wms_layers, query_layers, kml, is_single_tile, legend, '
        'legend_image, legend_rule, is_legend_expanded, min_resolution, max_resolution, '
        'disclaimer, identifier_attribute_field, exclude_properties, time_mode) '
        '(SELECT '
        'id, "isChecked" AS is_checked, icon, "layerType" AS layer_type, url, '
        '"imageType" AS image_type, style, dimensions, "matrixSet" AS matrix_set, '
        '"wmsUrl" AS wms_url, "wmsLayers" AS wms_layers, "queryLayers" AS query_layers, kml, '
        '"isSingleTile" AS is_single_tile, legend, "legendImage" AS legend_image, '
        '"legendRule" AS legend_rule, "isLegendExpanded" AS is_legend_expanded, '
        '"minResolution" AS min_resolution, "maxResolution" AS max_resolution, disclaimer, '
        '"identifierAttributeField" AS identifier_attribute_field, '
        '"excludeProperties" AS exclude_properties, "timeMode" AS time_mode '
        'FROM %(schema)s.layer)' % {'schema': schema}
    )

    op.drop_column('layer', 'isChecked', schema=schema)
    op.drop_column('layer', 'icon', schema=schema)
    op.drop_column('layer', 'layerType', schema=schema)
    op.drop_column('layer', 'url', schema=schema)
    op.drop_column('layer', 'imageType', schema=schema)
    op.drop_column('layer', 'style', schema=schema)
    op.drop_column('layer', 'dimensions', schema=schema)
    op.drop_column('layer', 'matrixSet', schema=schema)
    op.drop_column('layer', 'wmsUrl', schema=schema)
    op.drop_column('layer', 'wmsLayers', schema=schema)
    op.drop_column('layer', 'queryLayers', schema=schema)
    op.drop_column('layer', 'kml', schema=schema)
    op.drop_column('layer', 'isSingleTile', schema=schema)
    op.drop_column('layer', 'legend', schema=schema)
    op.drop_column('layer', 'legendImage', schema=schema)
    op.drop_column('layer', 'legendRule', schema=schema)
    op.drop_column('layer', 'isLegendExpanded', schema=schema)
    op.drop_column('layer', 'minResolution', schema=schema)
    op.drop_column('layer', 'maxResolution', schema=schema)
    op.drop_column('layer', 'disclaimer', schema=schema)
    op.drop_column('layer', 'identifierAttributeField', schema=schema)
    op.drop_column('layer', 'excludeProperties', schema=schema)
    op.drop_column('layer', 'timeMode', schema=schema)

    interface = Table(
        'interface', MetaData(),
        Column('name', Unicode),
        schema=schema,
    )
    op.bulk_insert(interface, [
        {'name': 'main'},
        {'name': 'mobile'},
        {'name': 'edit'},
        {'name': 'routing'},
    ])

    op.execute(
        'INSERT INTO %(schema)s.interface_layer (layer_id, interface_id) '
        '(SELECT l.id AS layer_id, i.id AS interface_id '
        'FROM %(schema)s.layer AS l, %(schema)s.interface AS i '
        'WHERE i.name in (\'main\', \'edit\', \'routing\') AND l."inDesktopViewer")' % {
            'schema': schema
        }
    )
    op.execute(
        'INSERT INTO %(schema)s.interface_layer (layer_id, interface_id) '
        '(SELECT l.id AS layer_id, i.id AS interface_id '
        'FROM %(schema)s.layer AS l, %(schema)s.interface AS i '
        'WHERE i.name = \'mobile\' AND l."inMobileViewer")' % {'schema': schema}
    )

    op.execute(
        'INSERT INTO %(schema)s.interface_theme (theme_id, interface_id) '
        '(SELECT l.id AS theme_id, i.id AS interface_id '
        'FROM %(schema)s.theme AS l, %(schema)s.interface AS i '
        'WHERE i.name in (\'main\', \'edit\', \'routing\') AND l."inDesktopViewer")' % {
            'schema': schema
        }
    )
    op.execute(
        'INSERT INTO %(schema)s.interface_theme (theme_id, interface_id) '
        '(SELECT l.id AS theme_id, i.id AS interface_id '
        'FROM %(schema)s.theme AS l, %(schema)s.interface AS i '
        'WHERE i.name = \'mobile\' AND l."inMobileViewer")' % {'schema': schema}
    )

    op.drop_column('layer', 'inMobileViewer', schema=schema)
    op.drop_column('layer', 'inDesktopViewer', schema=schema)

    op.alter_column('layer', 'geoTable', new_column_name='geo_table', schema=schema)

    op.drop_column('theme', 'inMobileViewer', schema=schema)
    op.drop_column('theme', 'inDesktopViewer', schema=schema)

    op.alter_column('treeitem', 'metadataURL', new_column_name='metadata_url', schema=schema)
    op.alter_column('layergroup', 'isExpanded', new_column_name='is_expanded', schema=schema)
    op.alter_column('layergroup', 'isInternalWMS', new_column_name='is_internal_wms', schema=schema)
    op.alter_column('layergroup', 'isBaseLayer', new_column_name='is_base_layer', schema=schema)

    op.create_table(
        'layer_internal_wms',
        Column(
            'id', Integer, ForeignKey(schema + '.layer.id'), primary_key=True
        ),
        Column('layer', Unicode),
        Column('image_type', Unicode(10)),
        Column('style', Unicode),
        Column('time_mode', Unicode(8)),
        schema=schema,
    )

    op.create_table(
        'layer_external_wms',
        Column(
            'id', Integer, ForeignKey(schema + '.layer.id'), primary_key=True
        ),
        Column('url', Unicode),
        Column('layer', Unicode),
        Column('image_type', Unicode(10)),
        Column('style', Unicode),
        Column('is_single_tile', Boolean),
        Column('time_mode', Unicode(8)),
        schema=schema,
    )

    op.create_table(
        'layer_wmts',
        Column(
            'id', Integer, ForeignKey(schema + '.layer.id'), primary_key=True,
        ),
        Column('url', Unicode),
        Column('layer', Unicode),
        Column('style', Unicode),
        Column('matrix_set', Unicode),
        schema=schema,
    )

    op.create_table(
        'ui_metadata',
        Column(
            'id', Integer, primary_key=True
        ),
        Column('name', Unicode),
        Column('value', Unicode),
        Column('description', Unicode),
        Column('item_id', Integer, ForeignKey(schema + '.treeitem.id'), nullable=False),
        schema=schema,
    )

    op.create_table(
        'wmts_dimension',
        Column(
            'id', Integer, primary_key=True
        ),
        Column('name', Unicode),
        Column('value', Unicode),
        Column('description', Unicode),
        Column('layer_id', Integer, ForeignKey(schema + '.layer_wmts.id'), nullable=False),
        schema=schema,
    )


def downgrade():
    schema = context.get_context().config.get_main_option('schema')

    op.drop_table('wmts_dimension', schema=schema)
    op.drop_table('ui_metadata', schema=schema)
    op.drop_table('layer_wmts', schema=schema)
    op.drop_table('layer_external_wms', schema=schema)
    op.drop_table('layer_internal_wms', schema=schema)

    op.add_column('layer', Column('inMobileViewer', Boolean, default=False), schema=schema)
    op.add_column('layer', Column('inDesktopViewer', Boolean, default=True), schema=schema)

    op.alter_column('layer', 'geo_table', new_column_name='geoTable', schema=schema)

    op.add_column('theme', Column('inMobileViewer', Boolean, default=False), schema=schema)
    op.add_column('theme', Column('inDesktopViewer', Boolean, default=True), schema=schema)

    op.alter_column('treeitem', 'metadata_url', new_column_name='metadataURL', schema=schema)
    op.alter_column('layergroup', 'is_expanded', new_column_name='isExpanded', schema=schema)
    op.alter_column('layergroup', 'is_internal_wms', new_column_name='isInternalWMS', schema=schema)
    op.alter_column('layergroup', 'is_base_layer', new_column_name='isBaseLayer', schema=schema)

    op.execute(
        'UPDATE ONLY %(schema)s.theme AS t '
        'SET "inDesktopViewer" = FALSE' % {'schema': schema}
    )
    op.execute(
        'UPDATE ONLY %(schema)s.layer AS t '
        'SET "inDesktopViewer" = FALSE' % {'schema': schema}
    )

    op.execute(
        'UPDATE ONLY %(schema)s.theme AS t '
        'SET "inMobileViewer" = TRUE '
        'FROM %(schema)s.interface AS i, %(schema)s.interface_theme AS it '
        'WHERE i.name = \'mobile\' AND i.id = it.interface_id AND it.theme_id = t.id' % {
            'schema': schema
        }
    )
    op.execute(
        'UPDATE ONLY %(schema)s.theme AS t '
        'SET "inDesktopViewer" = TRUE '
        'FROM %(schema)s.interface AS i, %(schema)s.interface_theme AS it '
        'WHERE i.name = \'main\' AND i.id = it.interface_id AND it.theme_id = t.id' % {
            'schema': schema
        }
    )
    op.execute(
        'UPDATE ONLY %(schema)s.layer AS l '
        'SET "inMobileViewer" = TRUE '
        'FROM %(schema)s.interface AS i, %(schema)s.interface_layer AS il '
        'WHERE i.name = \'mobile\' AND i.id = il.interface_id AND il.layer_id = l.id' % {
            'schema': schema
        }
    )
    op.execute(
        'UPDATE ONLY %(schema)s.layer AS l '
        'SET "inDesktopViewer" = TRUE '
        'FROM %(schema)s.interface AS i, %(schema)s.interface_layer AS il '
        'WHERE i.name = \'main\' AND i.id = il.interface_id AND il.layer_id = l.id' % {
            'schema': schema
        }
    )

    op.add_column('layer', Column('timeMode', Unicode(8)), schema=schema)
    op.add_column('layer', Column('excludeProperties', Unicode), schema=schema)
    op.add_column('layer', Column('identifierAttributeField', Unicode), schema=schema)
    op.add_column('layer', Column('disclaimer', Unicode), schema=schema)
    op.add_column('layer', Column('maxResolution', Float), schema=schema)
    op.add_column('layer', Column('minResolution', Float), schema=schema)
    op.add_column('layer', Column('isLegendExpanded', Boolean, default=False), schema=schema)
    op.add_column('layer', Column('legendRule', Unicode), schema=schema)
    op.add_column('layer', Column('legendImage', Unicode), schema=schema)
    op.add_column('layer', Column('legend', Boolean, default=True), schema=schema)
    op.add_column('layer', Column('isSingleTile', Boolean, default=False), schema=schema)
    op.add_column('layer', Column('kml', Unicode), schema=schema)
    op.add_column('layer', Column('queryLayers', Unicode), schema=schema)
    op.add_column('layer', Column('wmsLayers', Unicode), schema=schema)
    op.add_column('layer', Column('wmsUrl', Unicode), schema=schema)
    op.add_column('layer', Column('matrixSet', Unicode), schema=schema)
    op.add_column('layer', Column('dimensions', Unicode), schema=schema)
    op.add_column('layer', Column('style', Unicode), schema=schema)
    op.add_column('layer', Column('imageType', Unicode(10)), schema=schema)
    op.add_column('layer', Column('url', Unicode), schema=schema)
    op.add_column('layer', Column('layerType', Unicode(12)), schema=schema)
    op.add_column('layer', Column('icon', Unicode), schema=schema)
    op.add_column('layer', Column('isChecked', Boolean, default=True), schema=schema)

    op.execute(
        'UPDATE %(schema)s.layer AS l SET ('
        'id, "isChecked", icon, "layerType", url, "imageType", style, dimensions, "matrixSet", '
        '"wmsUrl", "wmsLayers", "queryLayers", kml, "isSingleTile", legend, "legendImage", '
        '"legendRule", "isLegendExpanded", "minResolution", "maxResolution", disclaimer, '
        '"identifierAttributeField", "excludeProperties", "timeMode"'
        ') = ('
        'o.id, o.is_checked, o.icon, o.layer_type, o.url, o.image_type, o.style, o.dimensions, '
        'o.matrix_set, o.wms_url, o.wms_layers, o.query_layers, o.kml, o.is_single_tile, '
        'o.legend, o.legend_image, o.legend_rule, o.is_legend_expanded, o.min_resolution, '
        'o.max_resolution, o.disclaimer, o.identifier_attribute_field, o.exclude_properties, '
        'o.time_mode '
        ') FROM %(schema)s.layerv1 AS o WHERE o.id = l.id' % {'schema': schema}
    )

    op.drop_table('layerv1', schema=schema)
    op.drop_table('interface_theme', schema=schema)
    op.drop_table('interface_layer', schema=schema)
    op.drop_table('interface', schema=schema)

    op.create_table(
        'user_functionality',
        Column(
            'user_id', Integer,
            ForeignKey(schema + '.user.id'), primary_key=True
        ),
        Column(
            'functionality_id', Integer,
            ForeignKey(schema + '.functionality.id'), primary_key=True
        ),
        schema=schema,
    )
